import numba as nb
import numpy as np

from vpso.typing import Array2d, Array3d


@nb.njit(
    nb.float64[:, :](
        nb.float64[:, :],  # x_best
        nb.bool_[:, :],  # mutation_mask
        nb.float64[:, :],  # lb
        nb.float64[:, :],  # ub
        nb.int32,  # nvec
        nb.int32,  # dim
        nb.types.NumPyRandomGeneratorType("NumPyRandomGeneratorType"),
    ),
    cache=True,
    nogil=True,
)
def polynomial_mutation(
    x_best: Array2d,
    mutation_mask: Array2d,
    lb: Array2d,
    ub: Array2d,
    nvec: int,
    dim: int,
    np_random: np.random.Generator,
) -> Array2d:
    """Performs a polynomial mutation on the best particle of each problem.

    Parameters
    ----------
    x_best : 2d array
        Best particles eligible for mutation. An array of shape `(nvec, dim)`.
    mutation_mask : 2d array
        Boolean mask indicating which particles should be mutated. An array of shape
        `(nvec, dim)`.
    lb : 2d array
        Lower bound of the search space. An array of shape `(nvec, dim)`.
    ub : 2d array
        Upper bound of the search space. An array of shape `(nvec, dim)`.
    nvec : int
        Number of vectorized problems.
    dim : int
        Dimensionality of the problem.
    np_random : np.random.Generator
        Random number generator.

    Returns
    -------
    2d array
        Modified best particles. An array of shape `(nvec, dim)`.
    """
    domain = ub - lb
    eta = np_random.uniform(6.0, 31.0, (nvec, np.int32(1)))
    mut_pow = 1.0 / eta
    xy1 = np.power((ub - x_best) / domain, eta)
    xy2 = np.power((x_best - lb) / domain, eta)
    R = np_random.random((nvec, dim))
    val1 = np.power(2.0 * R + (1.0 - 2.0 * R) * xy1, mut_pow)
    val2 = np.power(2.0 * (1.0 - R) + 2.0 * (R - 0.5) * xy2, mut_pow)
    deltaq = np.where(R <= 0.5, val1 - 1.0, 1.0 - val2)
    return np.where(mutation_mask, x_best + deltaq * domain, x_best).clip(lb, ub)


@nb.njit(
    nb.types.void(
        nb.float64[:, :, :],  # x
        nb.float64[:, :, :],  # px
        nb.float64[:, :],  # pf
        nb.float64[:, :, :],  # lb
        nb.float64[:, :, :],  # ub
        nb.int32,  # nvec
        nb.int32,  # dim
        nb.float64,  # mutation_prob
        nb.types.NumPyRandomGeneratorType("NumPyRandomGeneratorType"),
    ),
    cache=True,
    nogil=True,
)
def mutate(
    x: Array3d,
    px: Array3d,
    pf: Array2d,
    lb: Array3d,
    ub: Array3d,
    nvec: int,
    dim: int,
    mutation_prob: float,
    np_random: np.random.Generator,
) -> None:
    """Mutates in-place the best particle of each problem with a polynomial mutation.

    Parameters
    ----------
    x : 3d array
        Current positions of the particles. An array of shape `(N, M, d)`, where `N` is
        the number of vectorized problems to solve simultaneously, `M` is the number of
        particles in the swarm, and `d` is the dimension of the search space. This array
        is modified in-place.
    px : 3d array
        Best positions of the particles so far. An array of shape `(N, M, d)`.
    pf : 2d array
        Best values of the particles so far. An array of shape `(N, M)`.
    lb : 3d array
        Lower bound of the search space. An array of shape `(N, 1, d)`.
    ub : 3d array
        Upper bound of the search space. An array of shape `(N, 1, d)`.
    nvec : int
        Number of vectorized problems.
    dim : int
        Dimensionality of the problem.
    mutation_prob : float
        Probability of applying the mutation.
    np_random : np.random.Generator
        Random number generator.
    """
    # get the mutation mask for each vectorized problem, and for each dimension
    mutation_mask = np.logical_and(
        np_random.random((nvec, np.int32(1))) <= mutation_prob,
        np_random.random((nvec, dim)) <= min(0.5, 1 / dim),
    )
    if not mutation_mask.any():
        return

    # pick the best particle from each problem as target for mutation
    k = pf.argmin(1)
    # x_best = px[np.arange(nvec), k]  # cannot do this in njit
    x_best = np.empty((nvec, dim))
    for i, j in enumerate(k):
        x_best[i] = px[i, j]

    # compute mutation and apply it to the particle
    x_mutated = polynomial_mutation(
        x_best, mutation_mask, lb[:, 0], ub[:, 0], nvec, dim, np_random
    )
    for i, j in enumerate(k):
        x[i, j] = x_mutated[i]



import numba as nb
import numpy as np

from vpso.math import pso_equation
from vpso.typing import Array3d


@nb.njit(
    nb.types.UniTuple(nb.float64[:, :, :], 2)(
        nb.float64[:, :, :],  # x
        nb.float64[:, :, :],  # x_new
        nb.float64[:, :, :],  # v_new
        nb.float64[:, :, :],  # px
        nb.float64[:, :, :],  # sx
        nb.float64[:, :, :],  # v
        nb.float64[:, :, :],  # v_max
        nb.float64[:, :, :],  # lb
        nb.float64[:, :, :],  # ub
        nb.float64[:, :, :],  # w
        nb.float64[:, :, :],  # c1
        nb.float64[:, :, :],  # c2
        nb.int32,  # iters
        nb.types.NumPyRandomGeneratorType("NumPyRandomGeneratorType"),
    ),
    cache=True,
    nogil=True,
)
def repair_out_of_bounds(
    x: Array3d,
    x_new: Array3d,
    v_new: Array3d,
    px: Array3d,
    sx: Array3d,
    v: Array3d,
    v_max: Array3d,
    lb: Array3d,
    ub: Array3d,
    w: Array3d,
    c1: Array3d,
    c2: Array3d,
    iters: int,
    np_random: np.random.Generator,
) -> tuple[Array3d, Array3d]:
    """Repairs particles that have gone out of bounds with an iterative process and, if
    it failed, with random re-sampling.

    Parameters
    ----------
    x : 3d array
        Current positions of the particles. An array of shape `(N, M, d)`, where `N` is
        the number of vectorized problems to solve simultaneously, `M` is the number of
        particles in the swarm, and `d` is the dimension of the search space.
    x_new : 3d array
        New candidate positions of the particles. An array of shape `(N, M, d)`.
    v_new : 3d array
        new candidate velocities of the particles. An array of shape `(N, M, d)`.
    px : 3d array
        Best positions of the particles so far. An array of shape `(N, M, d)`.
    sx : 3d array
        Social best, i.e., the best particle so far. An array of shape `(N, 1, d)`.
    v : 3d array
        Current velocities of the particles. An array of shape `(N, M, d)`.
    v_max : 3d array
        Maximum velocities of the particles. An array of shape `(N, 1, d)`.
    lb : 3d array
        Lower bound of the search space. An array of shape `(N, 1, d)`.
    ub : 3d array
        Upper bound of the search space. An array of shape `(N, 1, d)`.
    w : 3d array
        Inertia weight. An array of shape `(N, 1, 1)`, where each element is used for
        the corresponding problem.
    c1 : 3d array
        Cognitive weight. An array of shape `(N, 1, 1)`, where each element is used for
        the corresponding problem.
    c2 : 3d array
        Social weight. An array of shape `(N, 1, 1)`, where each element is used for
        the corresponding problem.
    iters : int
        Number of iterations to try to repair the particles before random sampling.
    np_random : np.random.Generator
        Random number generator.

    Returns
    -------
    tuple of 3d arrays
        Repaired positions and velocities of the particles.
    """
    lmask = x_new < lb
    umask = x_new > ub
    any_lmask = lmask.any()
    any_umask = umask.any()
    if not any_lmask and not any_umask:
        return x_new, v_new

    for _ in range(iters):
        mask = np.logical_or(lmask, umask)
        x_new_, v_new_ = pso_equation(x, px, sx, v, v_max, w, c1, c2, np_random)
        x_new = np.where(mask, x_new_, x_new)
        v_new = np.where(mask, v_new_, v_new)
        lmask = x_new < lb
        umask = x_new > ub
        any_lmask = lmask.any()
        any_umask = umask.any()
        if not any_lmask and not any_umask:
            return x_new, v_new

    if any_lmask:
        x_new = np.where(lmask, lb + np_random.random(x.shape) * (x - lb), x_new)
    if any_umask:
        x_new = np.where(umask, ub - np_random.random(x.shape) * (ub - x), x_new)
    return x_new, v_new



import unittest

import numpy as np
from parameterized import parameterized

from vpso.reparation import repair_out_of_bounds


class TestReparation(unittest.TestCase):
    def test_repair_out_of_bounds__if_in_bounds__returns_immediately(self):
        np_random = np.random.default_rng()
        nvec, dim = np_random.integers(3, 10, size=2)
        swarmsize = np_random.integers(1000, 2000)
        ub = np.abs(np_random.normal(size=(nvec, 1, dim))) + 10
        lb = -np.abs(np_random.normal(size=(nvec, 1, dim))) - 10
        x = np_random.uniform(lb, ub, (nvec, swarmsize, dim))
        x_new = np_random.uniform(lb + 0.1, ub - 0.1, (nvec, swarmsize, dim))
        v_new = np_random.uniform(size=(nvec, swarmsize, dim))
        px = np_random.uniform(lb, ub, (nvec, swarmsize, dim))
        sx = np_random.uniform(lb, ub, (nvec, 1, dim))
        v = np_random.uniform(size=(nvec, swarmsize, dim))
        v_max = np_random.uniform(size=(nvec, 1, dim))
        w = np_random.uniform(size=(nvec, 1, 1))
        c1 = np_random.uniform(size=(nvec, 1, 1))
        c2 = np_random.uniform(size=(nvec, 1, 1))

        x_new_, v_new_ = repair_out_of_bounds(
            x,
            x_new,
            v_new,
            px,
            sx,
            v,
            v_max,
            lb,
            ub,
            w,
            c1,
            c2,
            200,
            np_random,
        )

        np.testing.assert_array_equal(x_new_, x_new)
        np.testing.assert_array_equal(v_new_, v_new)

    @parameterized.expand([(0,), (200,)])
    def test_repair_out_of_bounds__with_resamplimg_repair(self, iters: int):
        np_random = np.random.default_rng()
        nvec, dim = np_random.integers(3, 10, size=2)
        swarmsize = np_random.integers(1000, 2000)
        ub = np.abs(np_random.normal(size=(nvec, 1, dim))) + 10
        lb = -np.abs(np_random.normal(size=(nvec, 1, dim))) - 10
        x = np_random.uniform(lb, ub, (nvec, swarmsize, dim))
        x_new = np_random.uniform(lb - 10, ub + 10, (nvec, swarmsize, dim))
        v_new = np_random.uniform(size=(nvec, swarmsize, dim))
        px = np_random.uniform(lb, ub, (nvec, swarmsize, dim))
        sx = np_random.uniform(lb, ub, (nvec, 1, dim))
        v = np_random.uniform(size=(nvec, swarmsize, dim))
        v_max = np_random.uniform(size=(nvec, 1, dim))
        w = np_random.uniform(size=(nvec, 1, 1))
        c1 = np_random.uniform(size=(nvec, 1, 1))
        c2 = np_random.uniform(size=(nvec, 1, 1))

        x_new_, v_new_ = repair_out_of_bounds(
            x,
            x_new,
            v_new,
            px,
            sx,
            v,
            v_max,
            lb,
            ub,
            w,
            c1,
            c2,
            iters,
            np_random,
        )

        self.assertTrue(((x_new_ >= lb) & (x_new_ <= ub)).all())
        if iters > 0:
            mask = (x_new < lb) | (x_new > ub)  # only these velocities are modified
            self.assertTrue(((v_new_ >= -v_max) & (v_new_ <= v_max))[mask].all())


if __name__ == "__main__":
    unittest.main()



from typing import Callable, Union

import numpy as np
from numpy.typing import ArrayLike
from scipy.stats.qmc import LatinHypercube

from vpso.adaptation import adapt
from vpso.ask_and_tell import advance_population, generate_offsprings, get_best
from vpso.initialization import adjust_dimensions as adj_dim
from vpso.initialization import initialize_particles
from vpso.termination import termination
from vpso.typing import Array1d, Array1i, Array2d, Array3d


def vpso(
    func: Callable[[Array3d], ArrayLike],
    lb: Array2d,
    ub: Array2d,
    #
    swarmsize: int = 25,
    max_velocity_rate: Union[float, Array1d] = 0.2,
    w: Union[float, Array1d] = 0.9,
    c1: Union[float, Array1d] = 2.0,
    c2: Union[float, Array1d] = 2.0,
    #
    repair_iters: int = 20,
    perturb_best: bool = True,
    mutation_prob: float = 0.9,
    adaptive: bool = True,
    #
    maxiter: int = 400,
    maxevals: Union[int, float] = float("inf"),
    ftol: Union[float, Array1d] = 1e-9,
    xtol: Union[float, Array1d] = 1e-9,
    patience: Union[int, Array1i] = 30,
    #
    seed: Union[None, int, np.random.SeedSequence, np.random.Generator] = None,
) -> tuple[Array2d, Array1d, str]:
    """Vectorized Particle Swarm Optimization (VPSO). This implementation of PSO is able
    to solve multiple optimization problems simultaneously in a vectorized fashion. It
    runs the adaptive version of PSO based on [1].

    Parameters
    ----------
    func : callable
        The objective function to be minimized. Must take a 3d array of shape
        `(N, M, d)` as input, and return an output that can be converted to an array of
        shape `(N, M)`, where `N` is the number of vectorized problems to solve, `M` is
        the `swarmsize`, and `d` is the dimension of the search space.
    lb : 2d array
        Lower bound of the search space. An array of shape `(N, d)`. Can be different
        in each of the `N` vectorized problems.
    ub : 2d array
        Upper bound of the search space. An array of shape `(N, d)`.
    swarmsize : int, optional
        Number of particles in the swarm to solve each problem. By default, `25`.
    max_velocity_rate : float or array_like, optional
        Maximum velocity rate used to initialize the particles. By default, `0.2`. Can
        also be an 1d array_like of shape `(N,)` to specify a different value for each
        of the `N` vectorized problems.
    w : float or 1d array_like, optional
        Inertia weight. By default, `0.9`. Can
        also be an 1d array_like of shape `(N,)` to specify a different value for each
        of the `N` vectorized problems.
    c1 : float or 1d array_like, optional
        Cognitive weight. By default, `2.0`. Can also be an 1d array_like of shape
        `(N,)` to specify a different value for each of the `N` vectorized problems.
    c2 : float or 1d array_like, optional
        Social weight. By default, `2.0`. Can also be an 1d array_like of shape `(N,)`
        to specify a different value for each of the `N` vectorized problems.
    repair_iters : int, optional
        Number of iterations to repair particles that are outside bounds. If this
        reparation fails, the particle is randomly re-sampled. By default, `20`.
    perturb_best : bool, optional
        Whether to perturb the best particle in the swarm at each iteration.
        By default, `True`.
    mutation_prob : float, optional
        Probability of mutating the best particle in the swarm at each iteration. By
        default, `0.9`. Only used if `perturb_best=True`.
    adaptive : bool, optional
        Whether to adapt the weights at each iteration. By default, `True`.
    maxiter : int , optional
        Maximum number of iterations to run the optimization for. By default, `300`.
    maxevals : int or float(inf)
        Maximum number of evaluations of `f`, after which the algorithm is terminated.
        By default, `+inf`.
    ftol : float or 1d array_like of floats, optional
        Tolerance for changes in the objective function value before terminating the
        solver. Can also be an 1d array_like of shape `(N,)` to specify a different
        value for each of the `N` vectorized problems. By default, `1e-8`. Pass a
        negative value to disable this check.
    xtol : float or 1d array_like of floats, optional
        Tolerance for average changes in the objective minimizer before terminating the
        solver. Can also be an 1d array_like of shape `(N,)` to specify a different
        value for each of the `N` vectorized problems. By default, `1e-8`. Pass a
        negative value to disable this check.
    patience : int or 1d array_like of ints, optional
        Number of iterations to wait before terminating the solver if no improvement is
        witnessed. Can also be an 1d array_like of shape `(N,)` to specify a different
        value for each of the `N` vectorized problems. By default, `1`.
    seed : int or generator, optional
        Seed for the random number generator, or a generator. By default, `None`.

    Returns
    -------
    tuple of (2d array, 1d array, str)
        Returns a tuple containing
         - the best minimizer of each problem
         - the best minimum of each problem
         - the termination reason as a string

    References
    ----------
    [1] Z. H. Zhan, J. Zhang, Y. Li and H. S. H. Chung, "Adaptive Particle Swarm
        Optimization," in IEEE Transactions on Systems, Man, and Cybernetics, Part B
        (Cybernetics), vol. 39, no. 6, pp. 1362-1381, Dec. 2009,
        doi: 10.1109/TSMCB.2009.2015956.
    """
    # first, adjust some dimensions
    lb, ub, nvec, dim, max_velocity_rate, w, c1, c2, ftol, xtol, patience = adj_dim(
        lb, ub, max_velocity_rate, w, c1, c2, ftol, xtol, patience
    )

    # initialize particle positions and velocities
    np_random = np.random.default_rng(seed)
    lhs_sampler = LatinHypercube(d=nvec * dim, seed=np_random)
    x, v, v_max = initialize_particles(
        nvec, swarmsize, dim, lb, ub, max_velocity_rate, lhs_sampler, np_random
    )

    # initialize particle's best pos/value and global best
    px = x  # particle's best position
    pf = np.reshape(func(x), (nvec, swarmsize))  # particle's best value
    sx, sf = get_best(px, pf, nvec)  # social/global best position/value

    # main optimization loop
    patience_level = np.zeros((nvec, 2), dtype=np.int32)  # 2 level for x and for f
    evals = swarmsize
    termination_reason = "maxiter"
    for _ in range(maxiter):
        x, v = generate_offsprings(
            x,
            px,
            pf,
            sx,
            v,
            v_max,
            lb,
            ub,
            nvec,
            dim,
            w,
            c1,
            c2,
            repair_iters,
            perturb_best,
            mutation_prob,
            np_random,
        )
        f = np.reshape(func(x), (nvec, swarmsize))  # evaluate particles (non-jittable)
        px, pf = advance_population(x, f, px, pf)
        sx_new, sf_new = get_best(px, pf, nvec)

        # DEBUG
        # px.flags.writeable = (
        #     pf.flags.writeable
        # ) = sx.flags.writeable = sf.flags.writeable = False
        # assert (sf_new <= sf).all()

        if adaptive:
            w, c1, c2 = adapt(px, sx_new, nvec, swarmsize, lb, ub, w, c1, c2, np_random)

        # check termination conditions
        should_terminate, reason, range_min, range_max = termination(
            sx, sf, sx_new, sf_new, lb, ub, xtol, ftol, patience, patience_level
        )
        if should_terminate:
            termination_reason = f"{reason} ∈ [{range_min:e}, {range_max:e}]"
            break
        evals += swarmsize
        if evals >= maxevals:
            termination_reason = "maxevals"
            break

        # save new best
        sx = sx_new
        sf = sf_new
    return sx[:, 0], sf, termination_reason



import unittest

import numpy as np
from scipy.stats.qmc import LatinHypercube

from vpso.initialization import initialize_particles


class TestInitialization(unittest.TestCase):
    def test_initialize_particles(self):
        nvec, dim = np.random.randint(3, 10, size=2)
        swarmsize = np.random.randint(1000, 2000)
        ub = np.abs(np.random.randn(nvec, 1, dim))
        lb = -np.abs(np.random.randn(nvec, 1, dim))
        max_velocity_rate = np.random.uniform(0.1, 0.5)
        lhs_sampler = LatinHypercube(d=nvec * dim, scramble=False)
        np_random = np.random.default_rng()

        x, v, v_max = initialize_particles(
            nvec, swarmsize, dim, lb, ub, max_velocity_rate, lhs_sampler, np_random
        )
        self.assertTupleEqual(x.shape, (nvec, swarmsize, dim))
        self.assertTupleEqual(v.shape, (nvec, swarmsize, dim))
        self.assertTupleEqual(v_max.shape, (nvec, 1, dim))
        np.testing.assert_allclose(x.mean(1, keepdims=True), (ub + lb) / 2)
        self.assertTrue(((x >= lb) & (x <= ub)).all())
        self.assertTrue(((v >= 0) & (v <= v_max)).all())


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

from vpso import vpso_constrained


class TestNumericalConstrained(unittest.TestCase):
    def test_vpo_constrained(self):
        def func(x):
            return np.square(x[..., 0] - 2) + np.square(x[..., 1] - 1)

        def constraints(x):
            x_sq = np.square(x)
            g1a = x[..., 0] - 2 * x[..., 1] + 1
            g1b = -g1a
            g2 = x_sq[..., 0] / 4 + x_sq[..., 1] - 1
            return np.stack((g1a, g1b, g2), axis=-1)

        bounds = np.full((1, 2), 100.0)
        x_opt, _, _ = vpso_constrained(
            func=func,
            constraints=constraints,
            lb=-bounds,
            ub=+bounds,
            swarmsize=100,
            maxiter=1000,
            ftol=-1,  # disable ftol termination
            xtol=-1,  # disable xtol termination
            seed=1909,
            rapid_penalty=False,  # this function prefers a slower penalty increase
        )

        np.testing.assert_allclose(func(x_opt).item(), 1.393465, atol=1e-5, rtol=1e-5)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
from parameterized import parameterized
from scipy.spatial.distance import cdist, pdist, squareform

from vpso.math import batch_cdist, batch_cdist_and_pdist, batch_pdist, pso_equation


class TestMath(unittest.TestCase):
    @parameterized.expand((("euclidean",), ("sqeuclidean",)))
    def test_batch_cdist(self, dist_type: str):
        N, M1, M2, d = np.random.randint(10, 50, size=4)
        X = np.random.randn(N, M1, d)
        Y = np.random.randn(N, M2, d)
        actual = batch_cdist(X, Y, dist_type)
        expected = np.asarray([cdist(x, y, dist_type) for x, y in zip(X, Y)])
        np.testing.assert_allclose(actual, expected)

    @parameterized.expand((("euclidean",), ("sqeuclidean",)))
    def test_batch_pdist(self, dist_type: str):
        N, M, d = np.random.randint(10, 50, size=3)
        X = np.random.randn(N, M, d)
        actual = batch_pdist(X, dist_type)
        expected = np.asarray([squareform(pdist(x, dist_type)) for x in X])
        np.testing.assert_allclose(actual, expected)

    @parameterized.expand((("euclidean",), ("sqeuclidean",)))
    def test_batch_cdist_and_pdist(self, dist_type: str):
        N, M1, M2, d = np.random.randint(10, 50, size=4)
        X = np.random.randn(N, M1, d)
        Y = np.random.randn(N, M2, d)
        actual_c, actual_p = batch_cdist_and_pdist(X, Y, dist_type)
        expected_c = np.asarray([cdist(x, y, dist_type) for x, y in zip(X, Y)])
        expected_p = np.asarray([squareform(pdist(x, dist_type)) for x in X])
        np.testing.assert_allclose(actual_c, expected_c)
        np.testing.assert_allclose(actual_p, expected_p)

    def test_pso_equation(self):
        seed = np.random.randint(0, 1000)
        nvec, dim = np.random.randint(3, 10, size=2)
        swarmsize = np.random.randint(1000, 2000)
        np_random = np.random.default_rng(seed)
        r1 = np_random.uniform(size=(nvec, swarmsize, dim))
        r2 = np_random.uniform(size=(nvec, swarmsize, dim))
        x = np_random.normal(size=(nvec, swarmsize, dim))
        px = np_random.normal(size=(nvec, swarmsize, dim))
        sx = np_random.normal(size=(nvec, 1, dim))
        v = np_random.normal(size=(nvec, swarmsize, dim))
        v_max = np_random.normal(size=(nvec, 1, dim))
        w = np_random.uniform(size=(nvec, 1, 1))
        c1 = np_random.uniform(size=(nvec, 1, 1))
        c2 = np_random.uniform(size=(nvec, 1, 1))

        def original_implementation(X, P_X, S_X, V, V_max, w, c1, c2, r1, r2):
            inerta = w * V
            cognitive = c1 * r1 * (P_X - X)
            social = c2 * r2 * (S_X - X)
            Vp = inerta + cognitive + social
            Vp = Vp.clip(-V_max, V_max)
            Xp = X + Vp
            return Xp, Vp

        x_new_, v_new_ = [], []
        for i in range(nvec):
            o = original_implementation(
                x[i], px[i], sx[i], v[i], v_max[i], w[i], c1[i], c2[i], r1[i], r2[i]
            )
            x_new_.append(o[0])
            v_new_.append(o[1])
        x_new_, v_new_ = np.asarray(x_new_), np.asarray(v_new_)

        np_random = np.random.default_rng(seed)
        x_new, v_new = pso_equation(x, px, sx, v, v_max, w, c1, c2, np_random)

        np.testing.assert_allclose(x_new, x_new_)
        np.testing.assert_allclose(v_new, v_new_)


if __name__ == "__main__":
    unittest.main()



"""
Implements the adaptation mechanism of PSO described in [1].

References
----------
[1] Z. H. Zhan, J. Zhang, Y. Li and H. S. H. Chung, "Adaptive Particle Swarm
    Optimization," in IEEE Transactions on Systems, Man, and Cybernetics, Part B
    (Cybernetics), vol. 39, no. 6, pp. 1362-1381, Dec. 2009,
    doi: 10.1109/TSMCB.2009.2015956.
"""

import numba as nb
import numpy as np

from vpso.math import batch_cdist_and_pdist
from vpso.typing import Array1d, Array2d, Array3d


@nb.njit(
    nb.float64[:, :](nb.float64[:]),
    cache=True,
    nogil=True,
)
def adaptation_strategy(f: Array1d) -> Array2d:
    """Picks the adaptation strategy for each problem based on the ratio of average
    distances between particles and to the best particle.

    Parameters
    ----------
    f : 1d array
        Array of shape `N`.

    Returns
    -------
    2d array
        Array of shape `(N, 2)`. Each row contains the adaptation strategy for the
        `i`-th problem. The first column contains the adaptation strategy for `c1` and
        the second column contains the adaptation strategy for `c2`.
    """
    f = f[:, np.newaxis]  # add a dimension for broadcasting

    # NOTE: fails in numba's npython mode
    # deltas = np.full((f.size, 2), (-1.0, 1.0), dtype=f.dtype)  # initialize all to S4
    deltas = np.full((f.size, 2), -1.0)  # initialize all to S4 (fill in two steps)
    deltas[:, 1].fill(1.0)

    deltas = np.where(f <= 23 / 30, [(1.0, -1.0)], deltas)  # S1
    deltas = np.where(f < 0.5, [(0.5, -0.5)], deltas)  # S2
    deltas = np.where(f < 7 / 30, [(0.5, 0.5)], deltas)  # S3
    return deltas


@nb.njit(
    nb.types.UniTuple(nb.float64[:, :, :], 3)(
        nb.float64[:, :, :],  # px
        nb.float64[:, :, :],  # sx
        nb.int32,  # nvec
        nb.int32,  # swarmize
        nb.float64[:, :, :],  # lb
        nb.float64[:, :, :],  # ub
        nb.float64[:, :, :],  # w
        nb.float64[:, :, :],  # c1
        nb.float64[:, :, :],  # c2
        nb.types.NumPyRandomGeneratorType("NumPyRandomGeneratorType"),
    ),
    cache=True,
    nogil=True,
)
def adapt(
    px: Array3d,
    sx: Array3d,
    nvec: int,
    swarmsize: int,
    lb: Array3d,
    ub: Array3d,
    w: Array3d,
    c1: Array3d,
    c2: Array3d,
    np_random: np.random.Generator,
) -> tuple[Array3d, Array3d, Array3d]:
    """Performs the adaptation of the parameters `w`, `c1` and `c2` based on the
    stage of the algorithm.

    Parameters
    ----------
    px : 3d array
        Best positions of the particles so far. An array of shape `(N, M, d)`, where `N`
        is the number of vectorized problems to solve simultaneously, `M` the number of
        particle, and `d` is the dimension of the search space.
    sx : 3d array
        Social best, i.e., the best particle so far. An array of shape `(N, 1, d)`.
    nvec : int
        Number of vectorized problems.
    swarmsize : int
        Number of particles in the swarm.
    lb : 3d array
        Lower bound of the search space. An array of shape `(N, 1, d)`.
    ub : 3d array
        Upper bound of the search space. An array of shape `(N, 1, d)`.
    w : 3d array
        Inertia weight. An array of shape `(N, 1, 1)`, where each element is used for
        the corresponding problem.
    c1 : 3d array
        Cognitive weight. An array of shape `(N, 1, 1)`, where each element is used for
        the corresponding problem.
    c2 : 3d array
        Social weight. An array of shape `(N, 1, 1)`, where each element is used for
        the corresponding problem.
    np_random : np.random.Generator
        Random number generator.

    Returns
    -------
    tuple of 3d arrays
        The newly adapted parameters `w`, `c1` and `c2`.
    """
    domain = ub - lb
    G_, D_ = batch_cdist_and_pdist(px / domain, sx / domain, "euclidean")
    G = G_[:, :, 0].sum(1) / swarmsize
    D = D_.sum(2) / (swarmsize - 1)
    # NOTE: cannot run Dmin = D.min(1) and Dmax = D.max(1) in numba, so we use
    # https://stackoverflow.com/a/71214489/19648688
    Dmin = np.take_along_axis(D, D.argmin(1)[:, np.newaxis], 1)[:, 0]
    Dmax = np.take_along_axis(D, D.argmax(1)[:, np.newaxis], 1)[:, 0]
    stage = (G - Dmin) / (Dmax - Dmin + 1e-32)

    # adapt w
    w = (1 / (1 + 1.5 * np.exp(-2.6 * stage)))[:, np.newaxis, np.newaxis]

    # adapt c1 and c2
    deltas = adaptation_strategy(stage) * np_random.uniform(
        0.05, 0.1, size=(nvec, np.int32(1))
    )
    c1 = (c1 + deltas[:, 0, np.newaxis, np.newaxis]).clip(1.5, 2.5)
    c2 = (c2 + deltas[:, 1, np.newaxis, np.newaxis]).clip(1.5, 2.5)
    sum_c = c1 + c2
    mask = sum_c > 4
    c1 = np.where(mask, 4 * c1 / sum_c, c1)
    c2 = np.where(mask, 4 * c2 / sum_c, c2)
    return w, c1, c2



from typing import Literal

import numba as nb
import numpy as np

from vpso.typing import Array2d, Array3d


@nb.njit(
    nb.float64[:, :](nb.float64[:, :], nb.float64[:, :], nb.types.unicode_type),
    cache=True,
    nogil=True,
)
def cdist_func(
    x: Array2d, y: Array2d, type: Literal["euclidean", "sqeuclidean"]
) -> Array2d:
    """Computes the distance matrix for 2D arrays `x` and `y`.

    Parameters
    ----------
    x : 2d array
        A 2D array of shape `(N, d)`.
    y : 2d array
        A 2D array of shape `(M, d)`.
    type : {"euclidean", "sqeuclidean"}
        Type of distance to compute.

    Returns
    -------
    2d array
        Distance matrix of shape `(N, M)` between each pair of entries in `x` and `y`.
    """
    x2_sum = np.square(x).sum(axis=1)
    y2_sum = np.square(y).sum(axis=1)
    xy_dot = np.dot(x, y.T)
    D = np.maximum(x2_sum[:, np.newaxis] + y2_sum - 2 * xy_dot, 0.0)
    if type == "euclidean":
        return np.sqrt(D)
    if type == "sqeuclidean":
        return D
    raise ValueError(f"unknown cdist type: {type}")


@nb.njit(
    nb.float64[:, :](nb.float64[:, :], nb.types.unicode_type),
    cache=True,
    nogil=True,
)
def pdist_func(x: Array2d, type: Literal["euclidean", "sqeuclidean"]) -> Array2d:
    """Computes the pairwise distance matrix of the elements in the 2D array `x`.

    Parameters
    ----------
    x : 2d array
        A 2D array of shape `(N, d)`.
    type : {"euclidean", "sqeuclidean"}
        Type of distance to compute.

    Returns
    -------
    2d array
        Symmetric distance matrix of shape `(N, N)` between each pair of entries in `x`.
    """
    x2_sum = np.square(x).sum(axis=1)
    xx_dot = np.dot(x, x.T)
    D = np.maximum(x2_sum[:, np.newaxis] + x2_sum - 2 * xx_dot, 0.0)
    np.fill_diagonal(D, 0.0)
    if type == "euclidean":
        return np.sqrt(D)
    if type == "sqeuclidean":
        return D
    raise ValueError(f"unknown pdist type: {type}")


@nb.njit(
    nb.float64[:, :, :](
        nb.float64[:, :, :], nb.float64[:, :, :], nb.types.unicode_type
    ),
    cache=True,
    nogil=True,
    parallel=True,
)
def batch_cdist(
    X: Array3d, Y: Array3d, type: Literal["euclidean", "sqeuclidean"]
) -> Array3d:
    """Computes the distance matrices for 3D arrays.

    Parameters
    ----------
    X : 3d array
        An array of shape `(N, M, d)`.
    Y : 3d array
        An array of shape `(N, K, d)`.
    dist_func : callable, optional
        Distance function to use. By default, `scipy.spatial.distance.cdist_euclidean`
        is used. It must support the `out` argument.

    Returns
    -------
    3d array
        Distance matrices between each of the `(M, d)` and `(K, d)` matrices, where `d`
        is assumed to be the axis over which the distance is computed. The output has
        thus shape (N, M, K).
    """
    B, N, _ = X.shape
    out = np.empty((B, N, Y.shape[1]), dtype=X.dtype)
    for i in nb.prange(B):
        out[i] = cdist_func(X[i], Y[i], type)
    return out


@nb.njit(
    nb.float64[:, :, :](nb.float64[:, :, :], nb.types.unicode_type),
    cache=True,
    nogil=True,
    parallel=True,
)
def batch_pdist(X: Array3d, type: Literal["euclidean", "sqeuclidean"]) -> Array3d:
    """Computes the pairwise distance matrices for the entries of a 3D array.

    Parameters
    ----------
    X : 3d array
        An array of shape `(N, M, d)`.
    dist_func : callable, optional
        Distance function to use. By default, `scipy.spatial.distance.pdist_euclidean`
        is used. It must support the `out` argument.

    Returns
    -------
    3d array
        Distance matrix of shape `(N, M, M)` between each pair of entries of
        the `(M, d)` matrices, where `d` is assumed to be the axis over which the
        distance is computed.
    """
    B, N, _ = X.shape
    out = np.empty((B, N, N), dtype=X.dtype)
    for i in nb.prange(B):
        out[i] = pdist_func(X[i], type)
    return out


@nb.njit(
    nb.types.UniTuple(nb.float64[:, :, :], 2)(
        nb.float64[:, :, :], nb.float64[:, :, :], nb.types.unicode_type
    ),
    cache=True,
    nogil=True,
    parallel=True,
)
def batch_cdist_and_pdist(
    X: Array3d, Y: Array3d, type: Literal["euclidean", "sqeuclidean"]
) -> tuple[Array3d, Array3d]:
    """Computes the distance matrices between the 3D arrays `X` and `Y`, as well as the
    pairwise distance matrices for the entries of `X`.

    Parameters
    ----------
    X : 3d array
        An array of shape `(N, M, d)`.
    Y : 3d array
        An array of shape `(N, K, d)`.
    dist_func : callable, optional
        Distance function to use. By default, `scipy.spatial.distance.cdist_euclidean`
        is used. It must support the `out` argument.

    Returns
    -------
    tuple of 3d arrays
        Returns both distance matrices between `X` and `Y` and `X` itself. See
        `batch_cdist` and `batch_pdist` for more details.
    """
    B, N, _ = X.shape
    out_c = np.empty((B, N, Y.shape[1]), dtype=X.dtype)
    out_p = np.empty((B, N, N), dtype=X.dtype)
    for i in nb.prange(B):
        out_c[i] = cdist_func(X[i], Y[i], type)
        out_p[i] = pdist_func(X[i], type)
    return out_c, out_p


@nb.njit(
    nb.types.UniTuple(nb.float64[:, :, :], 2)(
        nb.float64[:, :, :],  # x
        nb.float64[:, :, :],  # px
        nb.float64[:, :, :],  # sx
        nb.float64[:, :, :],  # v
        nb.float64[:, :, :],  # v_max
        nb.float64[:, :, :],  # w
        nb.float64[:, :, :],  # c1
        nb.float64[:, :, :],  # c2
        nb.types.NumPyRandomGeneratorType("NumPyRandomGeneratorType"),
    ),
    cache=True,
    nogil=True,
)
def pso_equation(
    x: Array3d,
    px: Array3d,
    sx: Array3d,
    v: Array3d,
    v_max: Array3d,
    w: Array3d,
    c1: Array3d,
    c2: Array3d,
    np_random: np.random.Generator,
) -> tuple[Array3d, Array3d]:
    """Computes the new positions and velocities of particles in a PSO algorithm.

    Parameters
    ----------
    x : 3d array
        Current positions of the particles. An array of shape `(N, M, d)`, where `N` is
        the number of vectorized problems to solve simultaneously, `M` is the number of
        particles in the swarm, and `d` is the dimension of the search space.
    px : 3d array
        Best positions of the particles so far. An array of shape `(N, M, d)`.
    sx : 3d array
        Social best, i.e., the best particle so far. An array of shape `(N, 1, d)`.
    v : 3d array
        Current velocities of the particles. An array of shape `(N, M, d)`.
    v_max : 3d array
        Maximum velocities of the particles. An array of shape `(N, 1, d)`.
    w : 3d array
        Inertia weight. An array of shape `(N, 1, 1)`, where each element is used for
        the corresponding problem.
    c1 : 3d array
        Cognitive weight. An array of shape `(N, 1, 1)`, where each element is used for
        the corresponding problem.
    c2 : 3d array
        Social weight. An array of shape `(N, 1, 1)`, where each element is used for
        the corresponding problem.
    np_random : np.random.Generator
        Random number generator.

    Returns
    -------
    tuple of 3d arrays
        New positions and velocities of the particles.
    """
    r1 = np_random.random(x.shape)
    r2 = np_random.random(x.shape)

    inerta = w * v
    cognitive = c1 * r1 * (px - x)
    social = c2 * r2 * (sx - x)

    v_new = (inerta + cognitive + social).clip(-v_max, v_max)
    x_new = x + v_new
    return x_new, v_new  # type: ignore[return-value]  # thinks it's a bool array...



from numpy import float64, int32
from numpy.typing import NDArray
from typing_extensions import TypeAlias

Array1d: TypeAlias = NDArray[float64]
Array2d: TypeAlias = NDArray[float64]
Array3d: TypeAlias = NDArray[float64]
Array1i: TypeAlias = NDArray[int32]
Array2i: TypeAlias = NDArray[int32]



import unittest
from itertools import product

import numpy as np

from vpso.mutation import mutate, polynomial_mutation


class TestMutation(unittest.TestCase):
    def test_mutate__if_mutation_prob_is_zero__returns_immediately(self):
        np_random = np.random.default_rng()
        nvec, dim = np_random.integers(3, 10, size=2)
        swarmsize = np_random.integers(1000, 2000)
        ub = np.abs(np_random.normal(size=(nvec, 1, dim))) + 10
        lb = -np.abs(np_random.normal(size=(nvec, 1, dim))) - 10
        x_mutated = np_random.uniform(lb, ub, (nvec, swarmsize, dim))
        x_original = x_mutated.copy()
        px = np_random.uniform(lb, ub, (nvec, swarmsize, dim))
        pf = np_random.uniform(size=(nvec, swarmsize))
        mutation_prob = 0.0

        mutate(x_mutated, px, pf, lb, ub, nvec, dim, mutation_prob, np_random)
        np.testing.assert_array_equal(x_original, x_mutated)

    def test_mutate__performs_mutation(self):
        np_random = np.random.default_rng(17)
        nvec, dim = np_random.integers(3, 10, size=2)
        swarmsize = np_random.integers(1000, 2000)
        ub = np.abs(np_random.normal(size=(nvec, 1, dim))) + 10
        lb = -np.abs(np_random.normal(size=(nvec, 1, dim))) - 10
        x_mutated = np_random.uniform(lb, ub, (nvec, swarmsize, dim))
        x_original = x_mutated.copy()
        px = np_random.uniform(lb, ub, (nvec, swarmsize, dim))
        pf = np_random.uniform(size=(nvec, swarmsize))
        mutation_prob = 1.0

        mutate(x_mutated, px, pf, lb, ub, nvec, dim, mutation_prob, np_random)
        self.assertFalse(np.allclose(x_original, x_mutated))

        bests = set(pf.argmin(1))
        for i, j in product(range(nvec), range(swarmsize)):
            if j not in bests:
                np.testing.assert_array_equal(x_original[i, j], x_mutated[i, j])

    def test_polynomial_mutation(self):
        seed = np.random.randint(0, 1000)
        np_random = np.random.default_rng(seed)
        nvec, dim = np_random.integers(3, 10, size=2)
        ub = np.abs(np_random.normal(size=(nvec, dim))) + 10
        lb = -np.abs(np_random.normal(size=(nvec, dim))) - 10
        x = np_random.uniform(lb, ub, (nvec, dim))

        def original_implementation(X, eta, lb, ub, rand):
            delta1 = (X - lb) / (ub - lb)
            delta2 = (ub - X) / (ub - lb)
            mut_pow = 1.0 / (eta + 1.0)
            mask = rand <= 0.5
            mask_not = np.logical_not(mask)
            deltaq = np.zeros(X.shape)
            xy = 1.0 - delta1
            val = 2.0 * rand + (1.0 - 2.0 * rand) * (np.power(xy, (eta + 1.0)))
            d = np.power(val, mut_pow) - 1.0
            deltaq[mask] = d[mask]
            xy = 1.0 - delta2
            val = 2.0 * (1.0 - rand) + 2.0 * (rand - 0.5) * (np.power(xy, (eta + 1.0)))
            d = 1.0 - (np.power(val, mut_pow))
            deltaq[mask_not] = d[mask_not]
            return (X + deltaq * (ub - lb)).clip(lb, ub)

        np_random = np.random.default_rng(seed)
        eta = np_random.uniform(5, 30, size=nvec)
        rand = np_random.uniform(size=(nvec, dim))
        expected = [
            original_implementation(x[i], eta[i], lb[i], ub[i], rand[i])
            for i in range(nvec)
        ]

        np_random = np.random.default_rng(seed)
        actual = polynomial_mutation(
            x, np.full((nvec, dim), True), lb, ub, nvec, dim, np_random
        )
        np.testing.assert_allclose(actual, expected)


if __name__ == "__main__":
    unittest.main()



import unittest
from typing import Type

import numpy as np
from parameterized import parameterized
from pymoo.core.problem import Problem
from pymoo.problems.single import (
    Ackley,
    Griewank,
    Himmelblau,
    Rastrigin,
    Rosenbrock,
    Schwefel,
    Sphere,
    Zakharov,
)

from vpso import vpso

CLS = [Ackley, Griewank, Himmelblau, Rastrigin, Rosenbrock, Schwefel, Sphere, Zakharov]
np.set_printoptions(precision=3, suppress=True)


class TestNumerical(unittest.TestCase):
    @parameterized.expand([(cls,) for cls in CLS])
    def test_vpo(self, cls: Type[Problem]):
        np.random.seed(1909)
        nvec = np.random.randint(2, 10)
        swarmsize = 300
        # ftol = 1e-8
        maxiter = 500
        # period = 10
        problem = cls()

        def objs(x: np.ndarray) -> np.ndarray:
            return np.asarray(
                [problem.evaluate(x_, return_values_of=["F"]) for x_ in x]
            )

        actual_x, actual_f, _ = vpso(
            objs,
            np.tile(problem.xl, (nvec, 1)),
            np.tile(problem.xu, (nvec, 1)),
            #
            swarmsize=swarmsize,
            #
            maxiter=maxiter,
            #
            xtol=-1,
            ftol=-1,
            seed=np.random.randint(0, 1000),
        )

        pf = problem.pareto_front().item()
        ps = problem.pareto_set().squeeze()
        np.testing.assert_allclose(
            actual_f, pf, atol=1e-3, rtol=1e-3, err_msg=f"f {cls.__name__}"
        )
        if ps.ndim == 1:
            np.testing.assert_allclose(
                *np.broadcast_arrays(actual_x, ps),
                atol=1e-3,
                rtol=1e-3,
                err_msg=f"x {cls.__name__}",
            )
        else:
            for i in range(nvec):
                x_opt = actual_x[i]
                self.assertTrue(
                    any(np.allclose(x_opt, ps_, atol=1e-3, rtol=1e-3) for ps_ in ps),
                    msg=f"x {i} {cls.__name__}",
                )


if __name__ == "__main__":
    unittest.main()



from numbers import Number
from typing import Literal, Union

import numpy as np
from scipy.stats.qmc import LatinHypercube

from vpso.typing import Array1d, Array1i, Array2d, Array3d


def _asarray(val, nvec, dtype, ndim: Literal[1, 3]):
    if isinstance(val, Number):
        val = np.full(nvec, val, dtype=dtype)
    shape = (nvec, 1, 1) if ndim == 3 else (nvec,)
    return np.reshape(val, shape).astype(dtype, copy=False)


def adjust_dimensions(
    lb: Array2d,
    ub: Array2d,
    max_velocity_rate: Union[float, Array1d],
    w: Union[float, Array1d],
    c1: Union[float, Array1d],
    c2: Union[float, Array1d],
    ftol: Union[float, Array1d],
    xtol: Union[float, Array1d],
    patience: Union[int, Array1i],
) -> tuple[
    Array3d,
    Array3d,
    int,
    int,
    Array3d,
    Array3d,
    Array3d,
    Array3d,
    Array1d,
    Array1d,
    Array1i,
]:
    """Adjusts the dimensions of the input arrays to be compatible with the vectorized
    algorithm, i.e., adds dimensions when necessary or converts to array.

    lb : 2d array
        Lower bound of the search space. An array of shape `(N, d)`.
    ub : 2d array
        Upper bound of the search space. An array of shape `(N, d)`.
    max_velocity_rate : float or array, optional
        Maximum velocity rate used to initialize the particles. By default, `0.2`. Can
        also be an 1d array_like of shape `(N,)` to specify a different value for each
        of the `N` vectorized problems.
    w : float, optional
        Inertia weight. By default, `0.9`. Can
        also be an 1d array_like of shape `(N,)` to specify a different value for each
        of the `N` vectorized problems.
    c1 : float, optional
        Cognitive weight. By default, `2.0`. Can also be an 1d array_like of shape
        `(N,)` to specify a different value for each of the `N` vectorized problems.
    c2 : float, optional
        Social weight. By default, `2.0`. Can also be an 1d array_like of shape `(N,)`
        to specify a different value for each of the `N` vectorized problems.
    ftol : float or 1d array_like of floats, optional
        Tolerance for changes in the objective function value before terminating the
        solver. Can also be an 1d array_like of shape `(N,)` to specify a different
        value for each of the `N` vectorized problems. By default, `1e-8`.
    xtol : float or 1d array_like of floats, optional
        Tolerance for average changes in the objective minimizer before terminating the
        solver. Can also be an 1d array_like of shape `(N,)` to specify a different
        value for each of the `N` vectorized problems. By default, `1e-8`.
    patience : int or 1d array_like of ints, optional
        Number of iterations to wait before terminating the solver if no improvement is
        witnessed. Can also be an 1d array_like of shape `(N,)` to specify a different
        value for each of the `N` vectorized problems. By default, `1`.

    Returns
    -------
    tuple of 3d and 1d arrays
        Returns the `lb`, `ub`, `nvec`, `dim`, `max_velocity_rate`, `w`, `c1`, `c2`,
        `ftol`, `xtol`, and `patience`.
    """
    lb = lb[:, np.newaxis].astype(float, copy=False)  # add swarm dimension
    ub = ub[:, np.newaxis].astype(float, copy=False)  # add swarm dimension
    nvec, _, dim = lb.shape
    return (
        lb,
        ub,
        nvec,
        dim,
        _asarray(max_velocity_rate, nvec, np.float64, 3),
        _asarray(w, nvec, np.float64, 3),
        _asarray(c1, nvec, np.float64, 3),
        _asarray(c2, nvec, np.float64, 3),
        _asarray(ftol, nvec, np.float64, 1),
        _asarray(xtol, nvec, np.float64, 1),
        _asarray(patience, nvec, np.int32, 1),
    )


def initialize_particles(
    nvec: int,
    swarmsize: int,
    dim: int,
    lb: Array3d,
    ub: Array3d,
    max_velocity_rate: Array3d,
    lhs_sampler: LatinHypercube,
    np_random: np.random.Generator,
) -> tuple[Array3d, Array3d, Array3d]:
    """Initializes particle positions and velocities.

    Parameters
    ----------
    nvec : int
        Number of vectorized problems.
    swarmsize : int
        Size of the swarm for each problem.
    dim : int
        Dimensionality of the problem.
    lb : 3d array
        Lower bound of the search space. An array of shape `(N, 1, d)`.
    ub : 3d array
        Upper bound of the search space. An array of shape `(N, 1, d)`.
    max_velocity_rate : 3d array
        The maximum velocity rate (proportional to the search domain size) for each
        problem. An array of shape `(N, 1, 1)`.
    seed : int, optional
        Random seed.

    Returns
    -------
    tuple of 3d arrays
        Returns a tuple containing the particle positions, velocities, and maximum
        velocities.
    """
    domain = ub - lb

    x = lb + domain * lhs_sampler.random(swarmsize).reshape(
        swarmsize, nvec, dim
    ).transpose((1, 0, 2))

    v_max = max_velocity_rate * domain
    v = np_random.uniform(0, v_max, (nvec, swarmsize, dim))
    return x, v, v_max



import numba as nb
import numpy as np

from vpso.math import pso_equation
from vpso.mutation import mutate
from vpso.reparation import repair_out_of_bounds
from vpso.typing import Array1d, Array2d, Array3d


@nb.njit(
    nb.types.UniTuple(nb.float64[:, :, :], 2)(
        nb.float64[:, :, :],  # x
        nb.float64[:, :, :],  # px
        nb.float64[:, :],  # pf
        nb.float64[:, :, :],  # sx
        nb.float64[:, :, :],  # v
        nb.float64[:, :, :],  # v_max
        nb.float64[:, :, :],  # lb
        nb.float64[:, :, :],  # ub
        nb.int32,  # nvec
        nb.int32,  # dim
        nb.float64[:, :, :],  # w
        nb.float64[:, :, :],  # c1
        nb.float64[:, :, :],  # c2
        nb.int32,  # iters
        nb.bool_,  # perturb_best
        nb.float64,  # mutation_prob
        nb.types.NumPyRandomGeneratorType("NumPyRandomGeneratorType"),
    ),
    cache=True,
    nogil=True,
)
def generate_offsprings(
    x: Array3d,
    px: Array3d,
    pf: Array2d,
    sx: Array3d,
    v: Array3d,
    v_max: Array3d,
    lb: Array3d,
    ub: Array3d,
    nvec: int,
    dim: int,
    w: Array3d,
    c1: Array3d,
    c2: Array3d,
    repair_iters: int,
    perturb_best: bool,
    mutation_prob: float,
    np_random: np.random.Generator,
) -> tuple[Array3d, Array3d]:
    """Given the current swarm, generates the next generation of the particle swarm.

    Parameters
    ----------
    x : 3d array
        Current positions of the particles. An array of shape `(N, M, d)`, where `N` is
        the number of vectorized problems to solve simultaneously, `M` is the number of
        particles in the swarm, and `d` is the dimension of the search space.
    px : 3d array
        Best positions of the particles so far. An array of shape `(N, M, d)`.
    pf : 2d array
        Best values of the particles so far. An array of shape `(N, M)`.
    sx : 3d array
        Social best, i.e., the best particle so far. An array of shape `(N, 1, d)`.
    v : 3d array
        Current velocities of the particles. An array of shape `(N, M, d)`.
    v_max : 3d array
        Maximum velocities of the particles. An array of shape `(N, 1, d)`.
    lb : 3d array
        Lower bound of the search space. An array of shape `(N, 1, d)`.
    ub : 3d array
        Upper bound of the search space. An array of shape `(N, 1, d)`.
    nvec : int
        Number of vectorized problems.
    dim : int
        Dimensionality of the problem.
    w : 3d array
        Inertia weight. An array of shape `(N, 1, 1)`, where each element is used for
        the corresponding problem.
    c1 : 3d array
        Cognitive weight. An array of shape `(N, 1, 1)`, where each element is used for
        the corresponding problem.
    c2 : 3d array
        Social weight. An array of shape `(N, 1, 1)`, where each element is used for
        the corresponding problem.
    repair_iters : int
        Number of iterations to try to repair the particles before random sampling.
    perturb_best : bool
        Whether to perturb the best particle or not.
    mutation_prob : float
        Probability of applying the mutation.
    np_random : np.random.Generator
        Random number generator.

    Returns
    -------
    tuple of 3d arrays
        The positions and velocities of the particles in the next generation.
    """
    x_new, v_new = pso_equation(x, px, sx, v, v_max, w, c1, c2, np_random)
    x_new, v_new = repair_out_of_bounds(
        x, x_new, v_new, px, sx, v, v_max, lb, ub, w, c1, c2, repair_iters, np_random
    )
    if perturb_best:
        mutate(x_new, px, pf, lb, ub, nvec, dim, mutation_prob, np_random)
    return x_new, v_new


@nb.njit(
    nb.types.Tuple((nb.float64[:, :, :], nb.float64[:, :]))(
        nb.float64[:, :, :],  # x
        nb.float64[:, :],  # f
        nb.float64[:, :, :],  # px
        nb.float64[:, :],  # pf
    ),
    cache=True,
    nogil=True,
)
def advance_population(
    x: Array3d, f: Array2d, px: Array3d, pf: Array2d
) -> tuple[Array3d, Array2d]:
    """Advances the population by replacing the particles with better positions.

    Parameters
    ----------
    x : 3d array
        Current positions of the particles. An array of shape `(N, M, d)`, where `N` is
        the number of vectorized problems to solve simultaneously, `M` is the number of
        particles in the swarm, and `d` is the dimension of the search space.
    f : 2d array
        Current values of the particles. An array of shape `(N, M)`.
    px : 3d array
        Best positions of the particles so far. An array of shape `(N, M, d)`.
    pf : 2d array
        Best values of the particles so far. An array of shape `(N, M)`.

    Returns
    -------
    tuple of 3d and 2d arrays
        Returns the new positions and values of the best particles.
    """
    improvement_mask = f < pf
    px = np.where(improvement_mask[:, :, np.newaxis], x, px)
    pf = np.where(improvement_mask, f, pf)
    return px, pf


def get_best(px: Array3d, pf: Array2d, nvec: int) -> tuple[Array3d, Array1d]:
    """Returns the best particle and its value for each problem.

    Parameters
    ----------
    px : 3d array
        Best positions of the particles so far. An array of shape `(N, M, d)`, where `N`
        is the number of vectorized problems to solve simultaneously, `M` is the number
        of particles in the swarm, and `d` is the dimension of the search space.
    pf : 2d array
        Best values of the particles so far. An array of shape `(N, M)`.
    nvec : int
        Number of vectorized problems.

    Returns
    -------
    tuple of 3d and 1d arrays
        The best particle and its value for each problem with shape `(N, 1, d)` and
        `(N,)`, respectively.
    """
    idx = np.arange(nvec)
    k = pf.argmin(1)
    sx = px[idx, np.newaxis, k]  # (social/global) best particle
    sf = pf[idx, k]  # (social/global) best value
    return sx, sf



import unittest

import numpy as np
from pymoo.problems.functional import FunctionalProblem
from pymoo.util.misc import norm_eucl_dist

from vpso.adaptation import adapt, adaptation_strategy


def original_implementation1(f):
    def S1_explore(f):
        if f <= 0.4:
            return 0
        elif 0.4 < f <= 0.6:
            return 5 * f - 2
        elif 0.6 < f <= 0.7:
            return 1
        elif 0.7 < f <= 0.8:
            return -10 * f + 8
        elif 0.8 < f:
            return 0

    def S2_exploit(f):
        if f <= 0.2:
            return 0
        elif 0.2 < f <= 0.3:
            return 10 * f - 2
        elif 0.3 < f <= 0.4:
            return 1
        elif 0.4 < f <= 0.6:
            return -5 * f + 3
        elif 0.6 < f:
            return 0

    def S3_converge(f):
        if f <= 0.1:
            return 1
        elif 0.1 < f <= 0.3:
            return -5 * f + 1.5
        elif 0.3 < f:
            return 0

    def S4_jump_out(f):
        if f <= 0.7:
            return 0
        elif 0.7 < f <= 0.9:
            return 5 * f - 3.5
        elif 0.9 < f:
            return 1

    S = [S1_explore(f), S2_exploit(f), S3_converge(f), S4_jump_out(f)]
    strategy = np.argmax(S) + 1
    if strategy == 1:
        return 1, -1
    elif strategy == 2:
        return 0.5, -0.5
    elif strategy == 3:
        return 0.5, 0.5
    else:
        return -1, 1


def original_implementation2(problem, x, sx, w, c1, c2, delta):
    D = norm_eucl_dist(problem, x, x)
    mD = D.sum(axis=1) / (x.shape[0] - 1)
    _min, _max = mD.min(), mD.max()
    g_D = norm_eucl_dist(problem, sx, x).mean()
    f = (g_D - _min) / (_max - _min + 1e-32)
    strategy = original_implementation1(f)
    c1 = max(1.5, min(2.5, c1 + strategy[0] * delta))
    c2 = max(1.5, min(2.5, c2 + strategy[1] * delta))
    if c1 + c2 > 4.0:
        c1, c2 = 4.0 * (c1 / (c1 + c2)), 4.0 * (c2 / (c1 + c2))
    w = 1 / (1 + 1.5 * np.exp(-2.6 * f))
    return w, c1, c2


class TestAdaptation(unittest.TestCase):
    def test_adaptation_strategy(self):
        f = np.concatenate((np.arange(0, 1, 0.001), [0.23333, 0.5, 0.76666]))
        expected = [original_implementation1(f_) for f_ in f]
        actual = adaptation_strategy(f)
        np.testing.assert_array_equal(actual, expected)

    def test_adapt(self):
        nvec, dim = np.random.randint(3, 10, size=2)
        swarmsize = np.random.randint(100, 1000)
        ub = np.abs(np.random.rand(nvec, dim)) * 5 + 10
        lb = -np.abs(np.random.rand(nvec, dim)) * 5 - 10
        px = np.random.uniform(
            lb[:, np.newaxis], ub[:, np.newaxis], (nvec, swarmsize, dim)
        )
        sx = np.random.uniform(lb, ub, (nvec, dim))
        seed = np.random.randint(0, 1000)
        w = np.random.rand(nvec)
        c1 = np.random.rand(nvec) + 2
        c2 = np.random.rand(nvec) + 2

        np_random = np.random.default_rng(seed)
        deltas = 0.05 + np_random.random(size=nvec) * 0.05
        w_new, c1_new, c2_new = [], [], []
        for i in range(nvec):
            problem = FunctionalProblem(
                dim,
                lambda x: np.square(x).sum(),
                xl=lb[i],
                xu=ub[i],
            )
            o = original_implementation2(
                problem, px[i], sx[i], w[i], c1[i], c2[i], deltas[i]
            )
            w_new.append(o[0])
            c1_new.append(o[1])
            c2_new.append(o[2])

        np_random = np.random.default_rng(seed)
        w_new_, c1_new_, c2_new_ = adapt(
            px,
            sx[:, np.newaxis],
            nvec,
            swarmsize,
            lb[:, np.newaxis],
            ub[:, np.newaxis],
            w[:, np.newaxis, np.newaxis],
            c1[:, np.newaxis, np.newaxis],
            c2[:, np.newaxis, np.newaxis],
            np_random,
        )
        np.testing.assert_allclose(w_new_.squeeze(), w_new, err_msg="w")
        np.testing.assert_allclose(c1_new_.squeeze(), c1_new, err_msg="c1")
        np.testing.assert_allclose(c2_new_.squeeze(), c2_new, err_msg="c2")


if __name__ == "__main__":
    unittest.main()



from typing import Any, Callable

import numba as nb
import numpy as np
from numpy.typing import ArrayLike

from vpso import vpso
from vpso.typing import Array1d, Array2d, Array3d


@nb.njit(
    nb.float64[:, :](nb.int32, nb.float64[:, :, :], nb.boolean), cache=True, nogil=True
)
def penalty(k: int, g: Array3d, rapid: bool) -> Array2d:
    """Takes the current iteration and the constraints values at the swarm positions as
    input (N, M, d), and returns the penalty factor for each particle in each swarm,
    i.e., (N, M)."""
    q = np.maximum(0, g)
    theta = np.full(q.shape, 300.0)
    theta = np.where(q <= 1.0, 100.0, theta)
    theta = np.where(q <= 0.1, 20.0, theta)
    theta = np.where(q <= 0.001, 10.0, theta)
    gamma = np.where(q < 1, 1, 2)
    multiplier = np.sqrt(k)
    if rapid:
        multiplier *= k
    return multiplier * (theta * np.power(q, gamma)).sum(2)


class FuncWithPenalty:
    def __init__(
        self,
        f: Callable[[Array3d], ArrayLike],
        g: Callable[[Array3d], ArrayLike],
        rapid_penalty: bool = True,
    ) -> None:
        self.iteration = 0
        self.f = f
        self.g = g
        self.rapid = rapid_penalty

    def __call__(self, x: Array3d) -> Array2d:
        self.iteration += 1
        f = self.f(x)
        g = self.g(x)
        return f + penalty(self.iteration, g, self.rapid)


def vpso_constrained(
    func: Callable[[Array3d], ArrayLike],
    constraints: Callable[[Array3d], ArrayLike],
    *args: Any,
    rapid_penalty: bool = True,
    **kwargs: Any,
) -> tuple[Array2d, Array1d, str]:
    """Constrained variant of Vectorized Particle Swarm Optimization (VPSO), which can
    handle inequality constraints as in [1].

    Parameters
    ----------
    func : callable
        The objective function to be minimized. Same as in `vpso`.
    constraints : callable
        The inequality constraints to be satisfied, in the form `g(x) <= 0`. Must take a
        3d array of shape `(N, M, d)` as input, and return an output that can be
        converted to an array of shape `(N, M, g)`, where `N` is the number of
        vectorized problems to solve, `M` is the `swarmsize`, `d` is the dimension of
        the search space, and `g` is the  number of inequality constraints.
    rapid_penalty : bool, optional
        If `True`, the penalty factor is multiplied by the iteration number, increasing
        the speed at which the constraint penalty increases. Default is `True`.
    *args, **kwargs
        All other arguments are passed to `vpso`.

    Returns
    -------
    tuple of (2d array, 1d array, str)
        Returns a tuple containing
         - the best minimizer of each problem
         - the best minimum of each problem (including constraint penalty)
         - the termination reason as a string

    References
    ----------
    [1] Parsopoulos, K.E. and Vrahatis, M.N., 2002. Particle swarm optimization method
        for constrained optimization problems. Intelligent technologies–theory and
        application: New trends in intelligent technologies, 76(1), pp.214-220.
    """
    return vpso(FuncWithPenalty(func, constraints, rapid_penalty), *args, **kwargs)



import numba as nb
import numpy as np

from vpso.typing import Array1d, Array1i, Array2i, Array3d


@nb.njit(
    nb.types.UniTuple(nb.float64[:], 2)(
        nb.float64[:, :, :],  # sx
        nb.float64[:],  # sf
        nb.float64[:, :, :],  # sx_new
        nb.float64[:],  # sf_new
        nb.float64[:, :, :],  # lb
        nb.float64[:, :, :],  # ub
        nb.float64[:],  # xtol
        nb.float64[:],  # ftol
        nb.int32[:, :],  # current_patience_level
    ),
    cache=True,
    nogil=True,
)
def update_patience(
    sx: Array3d,
    sf: Array1d,
    sx_new: Array3d,
    sf_new: Array1d,
    lb: Array3d,
    ub: Array3d,
    xtol: Array1d,
    ftol: Array1d,
    current_patience_level: Array2i,
) -> tuple[Array1d, Array1d]:
    """Updates the patience level for each problem as a function of the previous and
    next best particle. The level is increased if tolerances are met, and reset to zero
    if the new best particle is better than the previous one by a margin larger than the
    tolerances. `current_patience_level` is modified in-place.

    Parameters
    ----------
    sx : 3d array
        Social best, i.e., the best particle so far. An array of shape `(N, 1, d)`,
        where `N` is the number of vectorized problems to solve simultaneously, and `d`
        is the dimension of the search space.
    sf : 1d array
        Social best value. An array of shape `(N,)`.
    sx_new : 3d array
        The new social best at the next iteration. An array of shape `(N, 1, d)`.
    sf_new : 1d array
        The new social best value at the next iteration. An array of shape `(N,)`.
    lb : 3d array
        Lower bound of the search space. An array of shape `(N, 1, d)`.
    ub : 3d array
        Upper bound of the search space. An array of shape `(N, 1, d)`.
    xtol : 1d array
        Tolerance for average changes in the objective minimizer before terminating the
        solver. An array of shape `(N,)`.
    ftol : 1d array
        Number of iterations to wait before terminating the solver if no improvement is
        witnessed. An array of shape `(N,)`.
    current_patience_level : 2d array of ints
        Current patience level, i.e., for how many iterations the tolerances have been
        met. An array of shape `(N, 2)`, where the first column corresponds to the
        tolerance in `x`, and the second column in `f`. This array is modified in-place.

    Returns
    -------
    tuple of 1d arrays
        Returns the current termination criteria for `x` and `f`, respectively, for each
        vectorized problem.
    """
    # NOTE: current patience level is to be modified in-place
    # normalize sx and sx_new and compute normalized euclidean distance (could call
    # batch_cdist here, but it's only one vector per bathc, so this should be faster)
    domain = ub - lb
    D = np.sqrt(np.square((sx - sx_new) / domain).sum(2))[:, 0]
    current_patience_level[:, 0] = np.where(
        D <= xtol, current_patience_level[:, 0] + 1, 0
    )

    # compute tolerance for f
    F = np.maximum(0.0, sf - sf_new)
    current_patience_level[:, 1] = np.where(
        F <= ftol, current_patience_level[:, 1] + 1, 0
    )
    return D, F


@nb.njit(
    nb.types.Tuple((nb.bool_, nb.types.unicode_type, nb.float64, nb.float64))(
        nb.float64[:, :, :],  # sx
        nb.float64[:],  # sf
        nb.float64[:, :, :],  # sx_new
        nb.float64[:],  # sf_new
        nb.float64[:, :, :],  # lb
        nb.float64[:, :, :],  # ub
        nb.float64[:],  # xtol
        nb.float64[:],  # ftol
        nb.int32[:],  # patience
        nb.int32[:, :],  # current_patience_level
    ),
    cache=True,
    nogil=True,
)
def termination(
    sx: Array3d,
    sf: Array1d,
    sx_new: Array3d,
    sf_new: Array1d,
    lb: Array3d,
    ub: Array3d,
    xtol: Array1d,
    ftol: Array1d,
    patience: Array1i,
    current_patience_level: Array2i,
) -> tuple[bool, str, float, float]:
    """Checks whether the solver should terminate, and the reason why. Updates the
    patience level for each problem as a function of the previous and next best
    particle. The level is increased if tolerances are met, and reset to zero if the new
    best particle is better than the previous one by a margin larger than the
    tolerances. `current_patience_level` is modified in-place.

    Parameters
    ----------
    sx : 3d array
        Social best, i.e., the best particle so far. An array of shape `(N, 1, d)`,
        where `N` is the number of vectorized problems to solve simultaneously, and `d`
        is the dimension of the search space.
    sf : 1d array
        Social best value. An array of shape `(N,)`.
    sx_new : 3d array
        The new social best at the next iteration. An array of shape `(N, 1, d)`.
    sf_new : 1d array
        The new social best value at the next iteration. An array of shape `(N,)`.
    lb : 3d array
        Lower bound of the search space. An array of shape `(N, 1, d)`.
    ub : 3d array
        Upper bound of the search space. An array of shape `(N, 1, d)`.
    xtol : 1d array
        Tolerance for average changes in the objective minimizer before terminating the
        solver. An array of shape `(N,)`.
    ftol : 1d array
        Number of iterations to wait before terminating the solver if no improvement is
        witnessed. An array of shape `(N,)`.
    patience : 1d array of ints, optional
        Number of iterations to wait before terminating the solver if no improvement is
        witnessed. An array of shape `(N,)`.
    current_patience_level : 2d array of ints
        Current patience level, i.e., for how many iterations the tolerances have been
        met. An array of shape `(N, 2)`, where the first column corresponds to the
        tolerance in `x`, and the second column in `f`. This array is modified in-place.

    Returns
    -------
    tuple of bool, str, range_min, range_max
        Returns whether the solver should terminate, and the reason why. The reason can
        be either `xtol` or `ftol`, depending on which tolerance was met first.
        Moreover, the min/max range for the tolerances that induced termination across
        the vectorized problems is returned.
    """
    D, F = update_patience(
        sx, sf, sx_new, sf_new, lb, ub, xtol, ftol, current_patience_level
    )
    # NOTE: can't f-string here because of numba, so we just return the ingredients
    if (current_patience_level[:, 0] >= patience).all():
        return True, "xtol", D.min(), D.max()
    if (current_patience_level[:, 1] >= patience).all():
        return True, "ftol", F.min(), F.max()
    return False, "", np.nan, np.nan
